# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import warnings
from copy import deepcopy
from collections import OrderedDict
from .spec import spec
from . import exceptions


# Module API

def preset(name):
    """https://github.com/frictionlessdata/goodtables-py#custom-presets
    """
    def decorator(func):
        registry.register_preset(func, name)
        return func
    return decorator


def check(name, type, context, position=None):
    """https://github.com/frictionlessdata/goodtables-py#custom-checks
    """
    def decorator(func):
        registry.register_check(func, name, type, context, position)
        return func
    return decorator


class Registry(object):

    # Public

    def __init__(self):
        self.__presets = OrderedDict()
        self.__checks = OrderedDict()

    def register_preset(self, func, name):
        self.__presets[name] = {
            'func': func,
            'name': name,
        }

    def compile_presets(self):
        return deepcopy(self.__presets)

    def register_check(self, func, name, type, context, position=None):
        check = {
            'func': func,
            'name': name,
            'type': type,
            'context': context,
        }

        # Validate check
        error = spec['errors'].get(name)
        if error and (error['type'] != type or error['context'] != context):
            message = 'Check "%s" is a part of the spec but type/context is incorrect'
            raise exceptions.GoodtablesException(message % name)
        elif not error and type != 'custom':
            message = 'Check "%s" is not a part of the spec should have type "custom"'
            raise exceptions.GoodtablesException(message % name)

        # Validate position
        if position:
            try:
                position = position.split(':', 1)
                assert position[0] in ['before', 'after']
                assert self.__checks.get(position[1])
            except (TypeError, AssertionError):
                message = 'Check "%s" has been registered at invalid position "%s"'
                raise exceptions.GoodtablesException(message % (name, position))

        # Insert into checks
        checks = OrderedDict()
        self.__checks.pop(name, None)
        for item_name, item_check in self.__checks.items():
            if position == 'before:%s' % item_name:
                checks[name] = check
            checks[item_name] = item_check
            if position == 'after:%s' % item_name:
                checks[name] = check
        if not position:
            checks[name] = check
        self.__checks = checks

    def compile_checks(self, include, exclude, **options):
        include = deepcopy(include)
        exclude = deepcopy(exclude)

        # Deprecated string checks
        if isinstance(include, six.string_types):
            warnings.warn(
                'Checks parameter as a string is deprecated. '
                'Please use a list',
                UserWarning)
            include = [include]

        # Deprecated dict checks
        if isinstance(include, dict):
            warnings.warn(
                'Checks parameter as a dict is deprecated. '
                'Please use a list',
                UserWarning)
            result_include = set()
            result_exclude = set()
            for name, enabled in include.items():
                if enabled:
                    result_include.add(name)
                else:
                    result_include.add('structure')
                    result_include.add('schema')
                    result_exclude.add(name)
            include = list(result_include)
            exclude = list(result_exclude)

        # Validate checks
        if not isinstance(include, list):
            message = 'Checks parameter "%s" is not valid' % include
            raise exceptions.GoodtablesException(message)

        # Expand checks
        for group in ['structure', 'schema']:
            for index, item in enumerate(list(include)):
                if item == group:
                    del include[index]
                    for code, error in spec['errors'].items():
                        # It's temporal skip
                        # https://github.com/frictionlessdata/goodtables-py/issues/174
                        if code == 'schema-error':
                            continue
                        if error['type'] == group:
                            include.append(code)

        # Compile checks
        compiled_checks = []
        for name, check in self.__checks.items():
            if name not in exclude:
                for item in include:
                    item_name = item
                    item_config = {}
                    if isinstance(item, dict):
                        item_name = list(item.keys())[0]
                        item_config = list(item.values())[0]
                    if item_name not in self.__checks:
                        message = 'Check "%s" is not registered'
                        raise exceptions.GoodtablesException(message % item_name)
                    if item_name == name:
                        compiled_check = deepcopy(check)
                        if isinstance(check['func'], type):
                            check_options = deepcopy(options)
                            check_options.update(item_config)
                            try:
                                compiled_check['func'] = check['func'](**check_options)
                            except Exception:
                                message = 'Check "%s" options "%s" error'
                                message = message % (check['name'], check_options)
                                raise exceptions.GoodtablesException(message)
                        compiled_checks.append(compiled_check)

        return compiled_checks


registry = Registry()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
from .inspector import Inspector


# Module API

def validate(source, **options):
    """https://github.com/frictionlessdata/goodtables-py#validate
    """

    # Extract settings
    settings = {}
    settings['checks'] = options.pop('checks', None)
    settings['skip_checks'] = options.pop('skip_checks', None)
    settings['infer_schema'] = options.pop('infer_schema', None)
    settings['infer_fields'] = options.pop('infer_fields', None)
    settings['order_fields'] = options.pop('order_fields', None)
    settings['error_limit'] = options.pop('error_limit', None)
    settings['table_limit'] = options.pop('table_limit', None)
    settings['row_limit'] = options.pop('row_limit', None)
    settings['custom_presets'] = options.pop('custom_presets', None)
    settings['custom_checks'] = options.pop('custom_checks', None)
    settings = {key: value for key, value in settings.items() if value is not None}

    # Support for pathlib.Path
    if hasattr(source, 'joinpath'):
        source = str(source)
    if isinstance(source, list):
        if source and isinstance(source[0], dict) and 'source' in source[0]:
            for index, item in enumerate(source):
                if hasattr(item['source'], 'joinpath'):
                    source[index]['source'] = str(item['source'])

    # Extract/infer preset
    preset = options.pop('preset', None)
    if preset is None:
        preset = 'table'
        if isinstance(source, six.string_types):
            if source.endswith('datapackage.json'):
                preset = 'datapackage'
        elif isinstance(source, dict):
            if 'resources' in source:
                preset = 'datapackage'
        elif isinstance(source, list):
            if source and isinstance(source[0], dict) and 'source' in source[0]:
                preset = 'nested'

    # Validate
    inspector = Inspector(**settings)
    report = inspector.inspect(source, preset=preset, **options)

    return report



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables.presets.datapackage import datapackage


# Preset

def test_preset_datapackage():
    warnings, tables = datapackage('data/datapackages/valid/datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


# Issues

def test_preset_datapackage_non_tabular_datapackage_issue_170():
    warnings, tables = datapackage('data/non_tabular_datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


def test_preset_datapackage_mixed_datapackage_issue_170():
    warnings, tables = datapackage('data/mixed_datapackage.json')
    assert len(warnings) == 0
    assert len(tables) == 2


def test_preset_datapackage_invalid_json_issue_192():
    warnings, tables = datapackage('data/invalid_json.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'Unable to parse JSON' in warnings[0]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..registry import check


# Module API

@check('unique-constraint', type='schema', context='body')
class UniqueConstraint(object):

    # Public

    def __init__(self, **options):
        self.__unique_fields_cache = None

    def check_row(self, errors, cells, row_number):

        # Prepare unique checks
        if self.__unique_fields_cache is None:
            self.__unique_fields_cache = _create_unique_fields_cache(cells)

        # Check unique
        for column_numbers, cache in self.__unique_fields_cache.items():
            values = tuple(cell.get('value')
                for column_number, cell in enumerate(cells, start=1)
                if column_number in column_numbers)
            if not all(map(lambda value: value is None, values)):
                if values in cache['data']:
                    message = spec['errors']['unique-constraint']['message']
                    message = message.format(
                        row_numbers=', '.join(map(str, cache['refs'] + [row_number])),
                        column_number=', '.join(map(str, column_numbers)))
                    errors.append({
                        'code': 'unique-constraint',
                        'message': message,
                        'row-number': row_number,
                        'column-number': column_numbers[0],
                    })
                cache['data'].add(values)
                cache['refs'].append(row_number)


# Internal

def _create_unique_fields_cache(cells):
    primary_key_column_numbers = []
    cache = {}

    # Unique
    for column_number, cell in enumerate(cells, start=1):
        if 'field' in cell:
            if cell['field'].descriptor.get('primaryKey'):
                primary_key_column_numbers.append(column_number)
            if cell['field'].constraints.get('unique'):
                cache[tuple([column_number])] = {
                    'data': set(),
                    'refs': [],
                }

    # Primary key
    if primary_key_column_numbers:
        cache[tuple(primary_key_column_numbers)] = {
            'data': set(),
            'refs': [],
        }

    return cache



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables import validate


# Validate

def test_check_sequential_value(log):
    source = [
        ['row', 'index2', 'index3'],
        [2, 1, 1],
        [3, 2, 3],
        [4, 3, 5],
        [5, 5, 6],
        [6],
    ]
    report = validate(source, checks=[
        {'sequential-value': {'column': 2}},
        {'sequential-value': {'column': 'index3'}},
    ])
    assert log(report) == [
        (1, 3, 3, 'sequential-value'),
        (1, 4, 3, 'sequential-value'),
        (1, 5, 2, 'sequential-value'),
        (1, 6, 2, 'sequential-value'),
        (1, 6, 3, 'sequential-value'),
    ]


def test_check_sequential_value_non_existent_column(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    report = validate(source, checks=[
        {'sequential-value': {'column': 3}},
        {'sequential-value': {'column': 'non-existent'}},
    ])
    assert log(report) == [
        (1, 2, None, 'sequential-value'),
        (1, 2, None, 'sequential-value'),
    ]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import click
import goodtables
import json as json_module
from pprint import pformat
click.disable_unicode_literals_warning = True


# Module API

@click.command()
@click.argument('source')
@click.option('--preset')
@click.option('--schema')
@click.option('--checks', type=lambda value: value.split(','))
@click.option('--skip-checks', type=lambda value: value.split(','))
@click.option('--infer-schema', is_flag=True)
@click.option('--infer-fields', is_flag=True)
@click.option('--order-fields', is_flag=True)
@click.option('--error-limit', type=int)
@click.option('--table-limit', type=int)
@click.option('--row-limit', type=int)
@click.option('--json', is_flag=True)
@click.version_option(goodtables.__version__, message='%(version)s')
def cli(source, json, **options):
    """https://github.com/frictionlessdata/goodtables-py#cli
    """
    options = {key: value for key, value in options.items() if value is not None}
    report = goodtables.validate(source, **options)
    _print_report(report, json=json)
    exit(not report['valid'])


# Internal
def _print_report(report, json=False):
    if json:
        return print(json_module.dumps(report, indent=4))
    color = 'green' if report['valid'] else 'red'
    tables = report.pop('tables')
    warnings = report.pop('warnings')
    click.secho('DATASET', bold=True)
    click.secho('='*7, bold=True)
    click.secho(pformat(report), fg=color, bold=True)
    if warnings:
        click.secho('-'*9, bold=True)
    for warning in warnings:
        click.secho('Warning: %s' % warning, fg='yellow')
    for table_number, table in enumerate(tables, start=1):
        click.secho('\nTABLE [%s]' % table_number, bold=True)
        click.secho('='*9, bold=True)
        color = 'green' if table['valid'] else 'red'
        errors = table.pop('errors')
        click.secho(pformat(table), fg=color, bold=True)
        if errors:
            click.secho('-'*9, bold=True)
        for error in errors:
            error = {key: value or '-' for key, value in error.items()}
            template = '[{row-number},{column-number}] [{code}] {message}'
            message = template.format(**error)
            click.secho(message)


# Main program

if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..spec import spec
from ..registry import check


# Module API

@check('extra-value', type='structure', context='body')
def extra_value(errors, cells, row_number):
    for cell in copy(cells):

        # Skip if header in cell
        if 'header' in cell:
            continue

        # Add error
        message = spec['errors']['extra-value']['message']
        message = message.format(
            row_number=row_number,
            column_number=cell['number'])
        errors.append({
            'code': 'extra-value',
            'message': message,
            'row-number': row_number,
            'column-number': cell['number'],
        })

        # Remove cell
        cells.remove(cell)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..spec import spec
from ..registry import check


# Module API

@check('required-constraint', type='schema', context='body')
def required_constraint(errors, cells, row_number):
    for cell in copy(cells):

        # Skip if cell is incomplete
        if not set(cell).issuperset(['number', 'header', 'field', 'value']):
            continue

        # Check constraint
        valid = cell['field'].test_value(cell['value'], constraints=['required'])
        if cell['field'].descriptor.get('primaryKey'):
            valid = valid and cell['field'].cast_value(cell['value']) is not None

        # Skip if valid
        if valid:
            continue

        # Add error
        message = spec['errors']['required-constraint']['message']
        message = message.format(
            row_number=row_number,
            column_number=cell['number'])
        errors.append({
            'code': 'required-constraint',
            'message': message,
            'row-number': row_number,
            'column-number': cell['number'],
        })

        # Remove cell
        cells.remove(cell)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from goodtables import validate, exceptions


# Validate

def test_check_deviated_value(log):
    source = [
        ['temperature'],
        [1],
        [-2],
        [7],
        [0],
        [1],
        [2],
        [5],
        [-4],
        [100],
        [8],
        [3],
    ]
    report = validate(source, checks=[
        {'deviated-value': {'column': 'temperature', 'average': 'median', 'interval': 3}},
    ])
    assert log(report) == [
        (1, 10, 1, 'deviated-value'),
    ]


def test_check_deviated_value_not_enough_data(log):
    source = [
        ['temperature'],
        [1],
    ]
    report = validate(source, checks=[
        {'deviated-value': {'column': 'temperature'}},
    ])
    assert log(report) == [
        (1, None, 1, 'deviated-value'),
    ]


def test_check_deviated_value_not_a_number(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    report = validate(source, checks=[
        {'deviated-value': {'column': 'name'}},
    ])
    assert log(report) == [
        (1, 2, 2, 'deviated-value'),
    ]


def test_check_deviated_value_non_existent_column(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    report = validate(source, checks=[
        {'deviated-value': {'column': 3}},
        {'deviated-value': {'column': 'non-existent'}},
    ])
    assert log(report) == [
        (1, 2, None, 'deviated-value'),
        (1, 2, None, 'deviated-value'),
    ]


def test_check_deviated_value_incorrect_average(log):
    source = [
        ['row', 'name'],
        [2, 'Alex'],
    ]
    with pytest.raises(exceptions.GoodtablesException):
        report = validate(source, checks=[
            {'deviated-value': {'column': 3, 'average': 'incorrect-average'}},
        ])



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from simpleeval import simple_eval
from ...registry import check


# Module API

@check('custom-constraint', type='custom', context='body')
class CustomConstraint(object):

    # Public

    def __init__(self, constraint, **options):
        self.__constraint = constraint

    def check_row(self, errors, cells, row_number):

        # Prepare names
        names = {}
        for cell in cells:
            if set(cell).issuperset(['header', 'value']):
                try:
                    names[cell['header']] = float(cell['value'])
                except ValueError:
                    pass

        # Check constraint
        try:
            # This call sould be considered as a safe expression evaluation
            # https://github.com/danthedeckie/simpleeval
            assert simple_eval(self.__constraint, names=names)
        except Exception:
            message = 'Custom constraint "%s" fails for row %s'
            return errors.append({
                'code': 'custom-constraint',
                'message': message % (self.__constraint, row_number),
                'row-number': row_number,
                'column-number': None,
            })



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ...registry import check


# Module API

@check('sequential-value', type='custom', context='body')
class SequentialValue(object):

    # Public

    def __init__(self, column, **options):
        self.__column = column
        self.__cursor = None

    def check_row(self, errors, cells, row_number):

        # Get cell
        cell = None
        for item in cells:
            if self.__column in [item['number'], item.get('header')]:
                cell = item
                break

        # Check cell
        if not cell:
            message = 'Sequential value check requires column "%s" to exist'
            return errors.append({
                'code': 'sequential-value',
                'message': message % self.__column,
                'row-number': row_number,
                'column-number': None,
            })

        # Get value
        try:
            value = int(cell.get('value'))
        except (TypeError, ValueError):
            message = 'Sequential value check requires column "%s" to be an integer'
            return errors.append({
                'code': 'sequential-value',
                'message': message % self.__column,
                'row-number': row_number,
                'column-number': cell['number'],
            })

        # Initiate cursor
        if self.__cursor is None:
            self.__cursor = value

        # Check value
        if self.__cursor != value:
            self.__cursor = value + 1
            message = 'Value "%s" is not a sequential in column %s for row %s'
            return errors.append({
                'code': 'sequential-value',
                'message': message % (value, cell['number'], row_number),
                'row-number': row_number,
                'column-number': cell['number'],
            })

        # Increment cursor
        self.__cursor += 1



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import json
from ..spec import spec
from ..registry import check


# Module API

@check('duplicate-row', type='structure', context='body')
class DuplicateRow(object):

    # Public

    def __init__(self, **options):
        self.__row_index = {}

    def check_row(self, errors, cells, row_number):

        # Get pointer
        try:
            pointer = hash(json.dumps(list(cell.get('value') for cell in cells)))
            references = self.__row_index.setdefault(pointer, [])
        except TypeError:
            pointer = None

        # Found pointer
        if pointer:

            # Add error
            if references:
                message = spec['errors']['duplicate-row']['message']
                message = message.format(
                    row_number=row_number,
                    row_numbers=', '.join(map(str, references)))
                errors.append({
                    'code': 'duplicate-row',
                    'message': message,
                    'row-number': row_number,
                    'column-number': None,
                })

            # Clear cells
            if references:
                del cells[:]

            # Update references
            references.append(row_number)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import statistics
from collections import OrderedDict
from ...registry import check
from ... import exceptions


# Module API

@check('deviated-value', type='custom', context='body')
class DeviatedValue(object):

    # Public

    def __init__(self, column, average='mean', interval=3, **options):

        # Set attributes
        self.__column = column
        self.__interval = interval
        self.__column_number = None
        self.__column_values = OrderedDict()
        self.__average_function = _AVERAGE_FUNCTIONS.get(average)

        # Validate average
        if not self.__average_function:
            message = 'Deviated value check supports only this average functions: %s'
            message = message % ', '.join(_AVERAGE_FUNCTIONS.keys())
            raise exceptions.GoodtablesError(message)

    def check_row(self, errors, cells, row_number):

        # Get cell
        cell = None
        for item in cells:
            if self.__column in [item['number'], item.get('header')]:
                cell = item
                break

        # Check cell
        if not cell:
            message = 'Deviated value check requires column "%s" to exist'
            return errors.append({
                'code': 'deviated-value',
                'message': message % self.__column,
                'row-number': row_number,
                'column-number': None,
            })

        # Get value
        try:
            value = float(cell['value'])
        except ValueError:
            message = 'Deviated value check requires column "%s" to be a number'
            return errors.append({
                'code': 'deviated-value',
                'message': message % self.__column,
                'row-number': row_number,
                'column-number': cell['number'],
            })

        # Collect value
        self.__column_values[row_number] = value
        self.__column_number = cell['number']

    def check_table(self, errors):

        # Skip if not values
        if not self.__column_values:
            return

        # Prepare interval
        try:
            stdev = statistics.stdev(self.__column_values.values())
            average = self.__average_function(self.__column_values.values())
            minimum = average - stdev * self.__interval
            maximum = average + stdev * self.__interval
        except Exception as exception:
            message = 'Deviated value check calculation issue: %s' % exception
            return errors.append({
                'code': 'deviated-value',
                'message': message,
                'row-number': None,
                'column-number': self.__column_number,
            })

        # Check values
        for row_number, value in self.__column_values.items():
            if not (minimum <= value <= maximum):
                message = 'Deviated value "%s" in column %s for row %s'
                errors.append({
                    'code': 'deviated-value',
                    'message': message % (value, self.__column_number, row_number),
                    'row-number': row_number,
                    'column-number': self.__column_number,
                })


# Internal

_AVERAGE_FUNCTIONS = {
    'mean': statistics.mean,
    'median': statistics.median,
    'mode': statistics.mode,
}



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
import tableschema
from ..spec import spec
from ..registry import check


# Module API

@check('type-or-format-error', type='schema', context='body')
def type_or_format_error(errors, cells, row_number):
    for cell in copy(cells):

        # Skip if cell is incomplete
        if not set(cell).issuperset(['number', 'header', 'field', 'value']):
            continue

        # Cast value
        try:
            valid = True
            cell['value'] = cell['field'].cast_value(cell['value'], constraints=False)
        except tableschema.exceptions.TableSchemaException:
            valid = False

        # Skip if valid
        if valid:
            continue

        # Add error
        message = spec['errors']['type-or-format-error']['message']
        message = message.format(
            value='"%s"' % cell['value'],
            row_number=row_number,
            column_number=cell['number'],
            field_type='"%s"' % cell['field'].type,
            field_format='"%s"' % cell['field'].format)
        errors.append({
            'code': 'type-or-format-error',
            'message': message,
            'row-number': row_number,
            'column-number': cell['number'],
        })

        # Remove cell
        cells.remove(cell)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import re
from copy import copy
from ..spec import spec
from ..registry import check


# Module API

@check('non-matching-header', type='schema', context='head')
class NonMatchingHeader(object):

    # Public

    def __init__(self, order_fields=False, **options):
        self.__order_fields = order_fields

    def check_headers(self, errors, cells, sample=None):
        if self.__order_fields:
            _check_with_ordering(errors, cells)
        else:
            _check_without_ordering(errors, cells)


# Internal


def _check_with_ordering(errors, cells):
    # Update fields order to maximally match headers order
    headers = [cell.get('header') for cell in cells]
    for index, header in enumerate(headers):
        if header is None:
            continue
        field_name = _get_field_name(cells[index])
        # Cell header and field_name are different
        if _slugify(header) != _slugify(field_name):
            # If there is the match in next cells - swap fields
            for cell in cells[index+1:]:
                # We've found field matching given header
                if _slugify(header) == _slugify(_get_field_name(cell)):
                    _swap_fields(cells[index], cell)
                    break
                if field_name:
                    # Given field matches some header also swap
                    if _slugify(field_name) == _slugify(cell.get('header')):
                        _swap_fields(cells[index], cell)
    # Run check with no field ordering
    _check_without_ordering(errors, cells)


def _check_without_ordering(errors, cells):
    for cell in copy(cells):
        if set(cell).issuperset(['number', 'header', 'field']):
            if cell['header'] != cell['field'].name:
                # Add error
                message = spec['errors']['non-matching-header']['message']
                message = message.format(
                    column_number=cell['number'],
                    field_name='"%s"' % cell['field'].name)
                errors.append({
                    'code': 'non-matching-header',
                    'message': message,
                    'row-number': None,
                    'column-number': cell['number'],
                })
                if _slugify(cell['header']) != _slugify(cell['field'].name):
                    # Remove cell
                    cells.remove(cell)


def _get_field_name(cell):
    field_name = None
    if 'field' in cell:
        field_name = cell['field'].name
    return field_name


def _slugify(string):
    if string is None:
        return None
    string = re.sub(r'[\W_]+', '', string)
    string = string.lower()
    return string


def _swap_fields(cell1, cell2):
    field1 = cell1.get('field')
    cell1['field'] = cell2.get('field')
    cell2['field'] = field1
    if cell1['field'] is None:
        del cell1['field']
    if cell2['field'] is None:
        del cell2['field']



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables import validate


# Validate

def test_check_truncated_value(log):
    source = [
        ['row', 'name', 'salary'],
        [2, 'a' * 255, 32766 ],
        [3, 'John', 32767],
        [4, 'Sam', 2147483647],
        [5],
    ]
    report = validate(source, checks=[
        'truncated-value',
    ])
    assert log(report) == [
        (1, 2, 2, 'truncated-value'),
        (1, 3, 3, 'truncated-value'),
        (1, 4, 3, 'truncated-value'),
    ]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..registry import check


# Module API

@check('blank-header', type='structure', context='head')
def blank_header(errors, cells, sample=None):
    for cell in cells:

        # Skip if cell have non blank header
        if cell.get('header', True):
            continue

        # Add error
        message = spec['errors']['blank-header']['message']
        message = message.format(column_number=cell['number'])
        errors.append({
            'code': 'blank-header',
            'message': message,
            'row-number': None,
            'column-number': cell['number'],
        })



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..registry import check


# Module API

@check('duplicate-header', type='structure', context='head')
def duplicate_header(errors, cells, sample=None):
    rindex = {}
    for cell in cells:

        # Skip if not header
        if 'header' not in cell:
            continue

        # Get references
        references = rindex.setdefault(cell['header'], [])

        # Add error
        if references:
            message = spec['errors']['duplicate-header']['message']
            message = message.format(
                column_number=cell['number'],
                column_numbers=', '.join(map(str, references)))
            errors.append({
                'code': 'duplicate-header',
                'message': message,
                'row-number': None,
                'column-number': cell['number'],
            })

        # Update references
        references.append(cell['number'])



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..registry import check


# Module API

@check('blank-row', type='structure', context='body')
def blank_row(errors, cells, row_number):
    if not list(filter(lambda cell: cell.get('value'), cells)):

        # Add error
        message = spec['errors']['blank-row']['message']
        message = message.format(row_number=row_number)
        errors.append({
            'code': 'blank-row',
            'message': message,
            'row-number': row_number,
            'column-number': None,
        })

        # Clear cells
        del cells[:]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..spec import spec
from ..registry import check


# Module API

@check('maximum-length-constraint', type='schema', context='body')
def maximum_length_constraint(errors, cells, row_number):
    for cell in cells:

        # Skip if cell is incomplete
        if not set(cell).issuperset(['number', 'header', 'field', 'value']):
            continue

        # Check constraint
        valid = cell['field'].test_value(cell['value'], constraints=['maxLength'])

        # Add error
        if not valid:
            message = spec['errors']['maximum-length-constraint']['message']
            message = message.format(
                value='"%s"' % cell['value'],
                row_number=row_number,
                column_number=cell['number'],
                constraint='"%s"' % cell['field'].constraints['maxLength'])
            errors.append({
                'code': 'maximum-length-constraint',
                'message': message,
                'row-number': row_number,
                'column-number': cell['number'],
            })



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import operator
import tabulator
from copy import copy
from tableschema import Schema
from six.moves import zip_longest
from multiprocessing.pool import ThreadPool
from .registry import registry
from . import exceptions
from . import config


# Module API

class Inspector(object):

    # Public

    def __init__(self,
                 checks=['structure', 'schema'],
                 skip_checks=[],
                 infer_schema=False,
                 infer_fields=False,
                 order_fields=False,
                 error_limit=config.DEFAULT_ERROR_LIMIT,
                 table_limit=config.DEFAULT_TABLE_LIMIT,
                 row_limit=config.DEFAULT_ROW_LIMIT):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Set attributes
        self.__checks = checks
        self.__skip_checks = skip_checks
        self.__infer_schema = infer_schema
        self.__infer_fields = infer_fields
        self.__order_fields = order_fields
        self.__error_limit = error_limit
        self.__table_limit = table_limit
        self.__row_limit = row_limit

    def inspect(self, source, preset='table', **options):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Start timer
        start = datetime.datetime.now()

        # Prepare preset
        try:
            presets = registry.compile_presets()
            preset_func = presets[preset]['func']
            if preset == 'nested':
                options['presets'] = presets
        except KeyError:
            message = 'Preset "%s" is not registered' % preset
            raise exceptions.GoodtablesException(message)

        # Prepare tables
        warnings, tables = preset_func(source, **options)
        if len(tables) > self.__table_limit:
            warnings.append(
                'Dataset inspection has reached %s table(s) limit' %
                (self.__table_limit))
            tables = tables[:self.__table_limit]

        # Collect table reports
        table_reports = []
        if tables:
            tasks = []
            pool = ThreadPool(processes=len(tables))
            for table in tables:
                tasks.append(pool.apply_async(self.__inspect_table, (table,)))
            for task in tasks:
                table_warnings, table_report = task.get()
                warnings.extend(table_warnings)
                table_reports.append(table_report)

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        report = {
            'time': round((stop - start).total_seconds(), 3),
            'valid': all(item['valid'] for item in table_reports),
            'error-count': sum(len(item['errors']) for item in table_reports),
            'table-count': len(tables),
            'tables': table_reports,
            'warnings': warnings,
            'preset': preset,
        }

        return report

    # Internal

    def __inspect_table(self, table):

        # Start timer
        start = datetime.datetime.now()

        # Prepare vars
        errors = []
        warnings = []
        headers = []
        row_number = 0
        fatal_error = False
        source = table['source']
        stream = table['stream']
        schema = table['schema']
        extra = table['extra']

        # Prepare checks
        checks = registry.compile_checks(
            table.get('checks', self.__checks), self.__skip_checks,
            order_fields=self.__order_fields, infer_fields=self.__infer_fields)

        # Prepare table
        try:
            stream.open()
            sample = stream.sample
            headers = stream.headers
            if headers is None:
                headers = [None] * len(sample[0]) if sample else []
            if _filter_checks(checks, type='schema'):
                if schema is None and self.__infer_schema:
                    schema = Schema()
                    schema.infer(sample, headers=headers)
            if schema is None:
                checks = _filter_checks(checks, type='schema', inverse=True)
        except Exception as exception:
            fatal_error = True
            error = _compose_error_from_exception(exception)
            errors.append(error)

        # Prepare schema
        if not fatal_error:
            if schema:
                if schema.primary_key:
                    for field in schema.descriptor.get('fields', []):
                        if field.get('name') in schema.primary_key:
                            field['primaryKey'] = True
                    schema.commit()
                for error in schema.errors:
                    fatal_error = True
                    error = _compose_error_from_schema_error(error)
                    errors.append(error)

        # Prepare cells
        if not fatal_error:
            cells = []
            fields = [None] * len(headers)
            if schema is not None:
                fields = schema.fields
            iterator = zip_longest(headers, fields, fillvalue=_FILLVALUE)
            for number, (header, field) in enumerate(iterator, start=1):
                cell = {'number': number}
                if header is not _FILLVALUE:
                    cell['header'] = header
                    cell['value'] = header
                if field is not _FILLVALUE:
                    cell['field'] = field
                cells.append(cell)

        # Head checks
        if not fatal_error:
            if None not in headers:
                head_checks = _filter_checks(checks, context='head')
                for check in head_checks:
                    if not cells:
                        break
                    check_func = getattr(check['func'], 'check_headers', check['func'])
                    check_func(errors, cells, sample)
                for error in errors:
                    error['row'] = None

        # Body checks
        if not fatal_error:
            cellmap = {cell['number']: cell for cell in cells}
            body_checks = _filter_checks(checks, context='body')
            with stream:
                extended_rows = stream.iter(extended=True)
                while True:
                    try:
                        row_number, _, row = next(extended_rows)
                    except StopIteration:
                        break
                    except Exception as exception:
                        fatal_error = True
                        error = _compose_error_from_exception(exception)
                        errors.append(error)
                        break
                    cells = []
                    iterator = zip_longest(headers, row, fillvalue=_FILLVALUE)
                    for number, (header, value) in enumerate(iterator, start=1):
                        cellref = cellmap.get(number, {})
                        cell = {'number': number}
                        if header is not _FILLVALUE:
                            cell['header'] = cellref.get('header', header)
                        if 'field' in cellref:
                            cell['field'] = cellref['field']
                        if value is not _FILLVALUE:
                            cell['value'] = value
                        cells.append(cell)
                    for check in body_checks:
                        if not cells:
                            break
                        check_func = getattr(check['func'], 'check_row', check['func'])
                        check_func(errors, cells, row_number)
                    for error in reversed(errors):
                        if 'row' in error:
                            break
                        error['row'] = row
                    if row_number >= self.__row_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s row(s) limit' %
                            (source, self.__row_limit))
                        break
                    if len(errors) >= self.__error_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s error(s) limit' %
                            (source, self.__error_limit))
                        break

        # Table checks
        if not fatal_error:
            for check in checks:
                check_func = getattr(check['func'], 'check_table', None)
                if check_func:
                    check_func(errors)

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        headers = headers if None not in headers else None
        errors = errors[:self.__error_limit]
        errors = _sort_errors(errors)
        report = copy(extra)
        report.update({
            'time': round((stop - start).total_seconds(), 3),
            'valid': not bool(errors),
            'error-count': len(errors),
            'row-count': row_number,
            'source': source,
            'headers': headers,
            'scheme': stream.scheme,
            'format': stream.format,
            'encoding': stream.encoding,
            'schema': 'table-schema' if schema else None,
            'errors': errors,
        })

        return warnings, report


# Internal

_FILLVALUE = '_FILLVALUE'


def _filter_checks(checks, type=None, context=None, inverse=False):
    result = []
    comparator = operator.eq if inverse else operator.ne
    for check in checks:
        if type and comparator(check['type'], type):
            continue
        if context and comparator(check['context'], context):
            continue
        result.append(check)
    return result


def _compose_error_from_exception(exception):
    code = 'source-error'
    message = str(exception)
    if isinstance(exception, tabulator.exceptions.SourceError):
        code = 'source-error'
    elif isinstance(exception, tabulator.exceptions.SchemeError):
        code = 'scheme-error'
    elif isinstance(exception, tabulator.exceptions.FormatError):
        code = 'format-error'
    elif isinstance(exception, tabulator.exceptions.EncodingError):
        code = 'encoding-error'
    elif isinstance(exception, tabulator.exceptions.IOError):
        code = 'io-error'
    elif isinstance(exception, tabulator.exceptions.HTTPError):
        code = 'http-error'
    return {
        'row': None,
        'code': code,
        'message': message,
        'row-number': None,
        'column-number': None,
    }


def _compose_error_from_schema_error(error):
    code = 'schema-error'
    message = 'Table Schema error: %s' % error
    return {
        'row': None,
        'code': code,
        'message': message,
        'row-number': None,
        'column-number': None,
    }


def _sort_errors(errors):
    return sorted(errors, key=lambda error:
        (error['row-number'] or 0, error['column-number'] or 0))



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from tableschema import Schema
from ..spec import spec
from ..registry import check


# Module API

@check('extra-header', type='schema', context='head')
class ExtraHeader(object):

    # Public
    def __init__(self, infer_fields=False, **options):
        self.__infer_fields = infer_fields

    def check_headers(self, errors, cells, sample):
        for cell in copy(cells):

            # Skip if cell has field
            if 'field' in cell:
                continue

            # Infer field
            if self.__infer_fields:
                column_sample = []
                for row in sample:
                    value = None
                    if len(row) >= cell['number']:
                        value = row[cell['number'] - 1]
                    column_sample.append(value)
                schema = Schema()
                schema.infer(column_sample, headers=[cell['header']])
                cell['field'] = schema.fields[0]

            # Add error/remove column
            else:
                message = spec['errors']['extra-header']['message']
                message = message.format(column_number=cell['number'])
                errors.append({
                    'code': 'extra-header',
                    'message': message,
                    'row-number': None,
                    'column-number': cell['number'],
                })
                cells.remove(cell)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
from ...registry import check


# Module API

@check('truncated-value', type='custom', context='body')
class TruncatedValue(object):

    # Public

    def __init__(self, **options):
        pass

    def check_row(self, errors, cells, row_number):
        for cell in cells:
            value = cell.get('value')
            truncated = False

            # Skip no value
            if not value:
                continue

            # Check string cutoff
            if isinstance(value, six.string_types):
                if len(value) in _TRUNCATED_STRING_LENGTHS:
                    truncated = True

            # Check integer cutoff
            try:
                value = int(value)
                if value in _TRUNCATED_INTEGER_VALUES:
                    truncated = True
            except ValueError:
                pass

            # Add error
            if truncated:
                message = 'Value in column %s for row %s is probably truncated'
                errors.append({
                    'code': 'truncated-value',
                    'message': message % (cell['number'], row_number),
                    'row-number': row_number,
                    'column-number': cell['number'],
                })


# Internal

_TRUNCATED_STRING_LENGTHS = [
    255,
]
_TRUNCATED_INTEGER_VALUES = [
    # BigInt
    18446744073709551616,
    9223372036854775807,
    # Int
    4294967295,
    2147483647,
    # SummedInt
    2097152,
    # SmallInt
    65535,
    32767,
]
